# Generated by Django 4.2.16 on 2024-12-04 11:56

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Delay',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('destination', models.EmailField(max_length=254)),
                ('delay_type', models.CharField(blank=True, max_length=50, null=True)),
                ('delay', models.JSONField()),
                ('mail', models.JSONField()),
                ('updated', models.DateTimeField(auto_now=True, verbose_name='Date updated')),
                ('created', models.DateTimeField(default=django.utils.timezone.now, verbose_name='Date created')),
                ('count', models.IntegerField(default=1)),
                ('user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='email_delays', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'ordering': ('-created',),
            },
        ),
        migrations.CreateModel(
            name='Complaint',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('destination', models.EmailField(max_length=254)),
                ('complaint_type', models.CharField(blank=True, max_length=50, null=True)),
                ('complaint', models.JSONField()),
                ('mail', models.JSONField()),
                ('updated', models.DateTimeField(auto_now=True, verbose_name='Date updated')),
                ('created', models.DateTimeField(default=django.utils.timezone.now, verbose_name='Date created')),
                ('count', models.IntegerField(default=1)),
                ('user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='email_complaints', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'ordering': ('-created',),
            },
        ),
        migrations.CreateModel(
            name='Bounce',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('destination', models.EmailField(max_length=254)),
                ('bounce_type', models.CharField(blank=True, max_length=50, null=True)),
                ('bounce_sub_type', models.CharField(blank=True, max_length=50, null=True)),
                ('bounce', models.JSONField()),
                ('mail', models.JSONField()),
                ('updated', models.DateTimeField(auto_now=True, verbose_name='Date updated')),
                ('created', models.DateTimeField(default=django.utils.timezone.now, verbose_name='Date created')),
                ('count', models.IntegerField(default=1)),
                ('user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='email_bounces', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'ordering': ('-created',),
            },
        ),
        migrations.AddConstraint(
            model_name='delay',
            constraint=models.UniqueConstraint(fields=('destination', 'delay_type'), name='unique_delay'),
        ),
        migrations.AddConstraint(
            model_name='complaint',
            constraint=models.UniqueConstraint(fields=('destination', 'complaint_type'), name='unique_complaint'),
        ),
        migrations.AddConstraint(
            model_name='bounce',
            constraint=models.UniqueConstraint(fields=('destination', 'bounce_type', 'bounce_sub_type'), name='unique_bounce'),
        ),
    ]



import re
from html.parser import HTMLParser as BaseHTMLParser


class HTMLParser(BaseHTMLParser):
    """
    Parser that strips all HTML tags from a string,
    removes the content of the <title> tag,
    and preserves the URLs in <a> tags.
    """
    def __init__(self):
        super().__init__(convert_charrefs=True)
        self._title = False
        self._href = None
        self._text = []

    def handle_data(self, data):
        if not self._title:
            text = data.strip()
            if len(text) > 0:
                text = re.sub('[ \t\r\n]+', ' ', text)
                self._text.append(text + ' ')

    def handle_starttag(self, tag, attrs):
        if tag == 'p':
            self._text.append('\n\n')
        elif tag == 'br':
            self._text.append('\n')
        elif tag == 'title':
            # recording being in title tag, so it's skipped in handle_data
            self._title = True
        elif tag == 'a':
            # attrs is a list of tuples: [('href', 'link'), ...]
            for attr in attrs:
                if attr[0] == 'href':
                    self._href = attr[1]
                    break

    def handle_endtag(self, tag):
        if tag == 'p':
            self._text.append('\n\n')
        elif tag == 'title':
            # end of title tag
            self._title = False
        elif tag == 'a' and self._href:
            # add the url that was found in starttag
            self._text.append(f'<{self._href}>')
            self._href = None

    def text(self):
        s = ''.join(self._text)

        # remove excessive white-space
        s = s.strip()
        s = re.sub(r'( {2,})', '', s)
        s = re.sub(r'(\s{3,})', '\n\n', s)
        return s



import re
import json
import base64
import logging
import requests

from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives import hashes
from cryptography import x509

from django.core.cache import cache

from django_aws_mail.config import mail_settings

logger = logging.getLogger(__name__)


class NotificationVerifier(object):
    """
    Class for validating Amazon SNS event notification messages as described here:
    https://docs.aws.amazon.com/sns/latest/dg/sns-verify-signature-of-message.html
    """
    required_notification_fields = {
        'Type', 'Message', 'Timestamp', 'Signature',
        'SignatureVersion', 'TopicArn', 'MessageId',
        'SigningCertURL'
    }

    allowed_notification_types = [
        'Notification', 'SubscriptionConfirmation', 'UnsubscribeConfirmation'
    ]

    def __init__(self, request):
        self._request = request
        self._verified = None
        self._notification = None
        self._message = None

    @property
    def is_verified(self):
        if self._verified is None:
            # checks must be run in this order
            self._verified = all([
                self.check_topic_header(),
                self.check_message_type_header(),
                self.check_notification(),
                self.check_keys(),
                self.check_type(),
                self.check_signature(),
                self.check_cert(),
            ])
        return self._verified

    def check_topic_header(self):
        # if necessary, check that the topic is correct
        if mail_settings.AWS_SNS_TOPIC_ARN:

            # confirm that the proper topic header was sent
            if 'HTTP_X_AMZ_SNS_TOPIC_ARN' not in self._request.META:
                logger.warning(f'Notification has no TopicArn header')
                return False

            # check to see if the topic is in the settings
            # bounces and complaints can come from multiple topics
            # MAIL_AWS_SNS_TOPIC_ARN is a list
            topic_hdr = self._request.META['HTTP_X_AMZ_SNS_TOPIC_ARN']
            if topic_hdr not in mail_settings.AWS_SNS_TOPIC_ARN:
                logger.warning(f"Notification contains bad topic: {topic_hdr}")
                return False

        return True

    def check_message_type_header(self):
        message_type_hdr = self._request.META['HTTP_X_AMZ_SNS_MESSAGE_TYPE']
        if message_type_hdr not in self.allowed_notification_types:
            logger.info(f"Unknown message type header {message_type_hdr}")
            return False
        return True

    def check_notification(self):
        """
        This check unpacks the notification from the request body and stores it locally!
        """
        request_body = self._request.body.decode('utf-8')
        try:
            self._notification = json.loads(request_body)
        except ValueError as e:
            logger.warning(f'Notification contains invalid JSON: {e}',
                           extra={'request': request_body})
            return False
        return True

    def check_keys(self):
        # ensure that the notification contains all the expected keys
        if not self.required_notification_fields <= self._notification.keys():
            logger.warning('Notification is missing required keys')
            return False
        return True

    def check_type(self):
        # ensure that the type of notification is one we'll accept
        if self._notification.get('Type') not in self.allowed_notification_types:
            logger.info(f"Unknown notification type {self._notification.get('Type')}")
            return False
        return True

    def check_signature(self):
        if self._notification.get('SignatureVersion') != '1':
            logger.warning('Invalid signature version. Unable to verify signature.')
            return False
        return True

    def check_cert(self):
        cert_url = self._notification.get('SigningCertURL')

        # confirm that the signing certificate is hosted on a correct domain
        # AWS by default uses sns.{region}.amazonaws.com
        pattern = r'^https://sns\.[-a-z0-9]+\.amazonaws\.com/'
        if not re.search(pattern, cert_url):
            logger.warning(f'Invalid certificate URL: {cert_url}')
            return False

        # verify that the notification is signed by Amazon
        # if getattr(settings, 'MAIL_AWS_SNS_VERIFY_CERTIFICATE', True):
        if mail_settings.AWS_SNS_VERIFY_CERTIFICATE:
            # get certificate
            pem_data = self.get_keyfile(cert_url)
            if not pem_data:
                return False
            cert = x509.load_pem_x509_certificate(pem_data)

            # get decoded signature
            signature = self._notification.get('Signature')
            if not signature:
                return False
            signature = bytes(base64.b64decode(signature))

            # retrieve canonical message data
            message = self.get_canonical_message()
            if not message:
                return False

            # figure out hash algorythm based on signature version
            signature_version = self._notification['SignatureVersion']
            if signature_version == '1':
                hash_algo = hashes.SHA1()
            elif signature_version == '2':
                hash_algo = hashes.SHA256()
            else:
                return False

            # verify message using cert public key and signature
            public_key = cert.public_key()
            try:
                public_key.verify(signature, message, padding.PKCS1v15(), hash_algo)
            except InvalidSignature:
                logger.error('Notification certificate verification failure')
                return False

        return True

    def check_message(self):
        """
        This check unpacks the message from the notification and stores it locally!
        This check is _not_ run with the other checks, so a http 200 can be returned.
        """
        try:
            self._message = json.loads(self._notification.get('Message'))
        except ValueError as e:
            # This message is not JSON. But we need to return a 200 status code
            # so that Amazon doesn't attempt to deliver the message again
            logger.warning(f'Invalid JSON Message Received: {e}',
                           extra={'notification': self._notification})
            return False
        return True

    def get_canonical_message(self):
        """
        Retrieve canonical message from Amazon SNS message, based on the notification type.
        :return: message as bytes
        """
        notification_type = self._notification.get('Type')
        if notification_type == 'Notification':
            if "Subject" in self._notification:
                fields = ['Message', 'MessageId', 'Subject', 'Timestamp', 'TopicArn', 'Type']
            else:
                fields = ['Message', 'MessageId', 'Timestamp', 'TopicArn', 'Type']
        elif notification_type in ['SubscriptionConfirmation', 'UnsubscribeConfirmation']:
            fields = ['Message', 'MessageId', 'SubscribeURL', 'Timestamp', 'Token', 'TopicArn', 'Type']
        else:
            logger.warning(f'Unrecognized SNS notification type: {notification_type}')
            return None

        pairs = [f'{field}\n{self._notification.get(field)}' for field in fields]
        message = '\n'.join(pairs) + '\n'
        return bytes(message, 'utf-8')

    @staticmethod
    def get_keyfile(cert_url):
        """
        Function to acquire the keyfile

        SNS keys expire and Amazon does not promise they will use the same key
        for all SNS requests. So we need to keep a copy of the cert in our cache.
        """
        pem_data = cache.get(cert_url)
        if not pem_data:
            try:
                pem_data = requests.get(cert_url).content
            except OSError:
                logger.error(f'Unable to retrieve certificate, URL: {cert_url}')
                return None

            cache.set(cert_url, pem_data)

        return pem_data

    def get_notification(self):
        if not self._notification:
            _ = self.check_notification()
        return self._notification

    def get_message(self):
        if not self._message:
            _ = self.check_message()
        return self._message



"""
Webhook for handling AWS SNS event messages.
Inspired by django-bouncy and django-ses.
"""
import re
import logging
from urllib.error import URLError
from urllib.parse import urlparse
from urllib.request import urlopen

from django.http import HttpResponseBadRequest, HttpResponse, Http404
from django.views.generic import View
from django.views.decorators.csrf import csrf_exempt
from django.utils.decorators import method_decorator

from django_aws_mail import signals
from django_aws_mail.config import mail_settings
from django_aws_mail.verifier import NotificationVerifier

logger = logging.getLogger(__name__)


@method_decorator(csrf_exempt, name='dispatch')
class AwsSnsWebhook(View):
    http_method_names = ['post']

    def http_method_not_allowed(self, request, *args, **kwargs):
        # hide this endpoint, non-POST requests will return an HTTP 404
        raise Http404

    def post(self, request, *args, **kwargs):
        verifier = NotificationVerifier(request)
        if mail_settings.AWS_SNS_VERIFY_NOTIFICATION and not verifier.is_verified:
            return HttpResponseBadRequest('Invalid notification')

        notification = verifier.get_notification()
        if not notification:
            return HttpResponseBadRequest('Invalid notification')

        notification_type = notification.get('Type')
        if notification_type == 'SubscriptionConfirmation':
            self.handle_subscription_confirmation(notification)

        elif notification_type == 'UnsubscribeConfirmation':
            self.handle_unsubscribe_confirmation(notification)

        elif notification_type == 'Notification':
            message = verifier.get_message()
            if message:
                # handle event types
                event_type = message.get('eventType')
                if event_type == 'Bounce':
                    self.handle_bounce(message)
                elif event_type == 'Complaint':
                    self.handle_complaint(message)
                elif event_type == 'Delivery':
                    self.handle_delivery(message)
                elif event_type == 'DeliveryDelay':
                    self.handle_delivery_delay(message)
                elif event_type == 'Send':
                    self.handle_send(message)
                elif event_type == 'Reject':
                    self.handle_reject(message)
                elif event_type == 'Open':
                    self.handle_open(message)
                elif event_type == 'Click':
                    self.handle_click(message)
                else:
                    self.handle_unknown_event(message)

        else:
            self.handle_unknown_notification(notification)

        return HttpResponse('ok')

    def handle_unknown_notification(self, notification):
        logger.warning(f"Received unknown notification type: {notification.get('Type')}",
                       extra={'notification': notification})

    def handle_unknown_event(self, message):
        logger.warning(f"Received unknown event type: {message.get('eventType')}",
                       extra={'message': message})

    def handle_subscription_confirmation(self, notification):
        logger.info((f"Received subscription confirmation, "
                     f"TopicArn: {notification.get('TopicArn')}"),
                    extra={'notification': notification})

        # check subscribe url
        url = notification.get('SubscribeURL')
        domain = urlparse(url).netloc
        pattern = r"sns.[a-z0-9\-]+.amazonaws.com$"
        if not re.search(pattern, domain):
            logger.error(f'Invalid subscription domain {url}')
            return HttpResponseBadRequest('Improper subscription domain')

        # simply visit url
        try:
            _ = urlopen(url).read()
            logger.info(f'Subscription confirmation request sent, URL: {url}')
        except URLError as e:
            logger.error(f'HTTP error creating subscription {e.reason}',
                         extra={'notification': notification})

    def handle_unsubscribe_confirmation(self, notification):
        # log message, no need to visit any url
        logger.info((f"Received unsubscription confirmation, "
                     f"TopicArn: {notification.get('TopicArn')}"),
                    extra={'notification': notification})

    def handle_bounce(self, message):
        self.handle_event('bounce', signals.mail_bounce, message)

    def handle_complaint(self, message):
        self.handle_event('complaint', signals.mail_complaint, message)

    def handle_delivery(self, message):
        self.handle_event('delivery', signals.mail_delivery, message)

    def handle_delivery_delay(self, message):
        self.handle_event('deliveryDelay', signals.mail_delivery_delay, message)

    def handle_send(self, message):
        self.handle_event('send', signals.mail_send, message)

    def handle_reject(self, message):
        self.handle_event('reject', signals.mail_reject, message)

    def handle_open(self, message):
        self.handle_event('open', signals.mail_open, message)

    def handle_click(self, message):
        self.handle_event('click', signals.mail_click, message)

    def handle_event(self, event_name, signal, message):
        mail = message.get('mail')
        event = message.get(event_name, {})

        logger.info(f"Received email {event_name} event for {mail['destination'][0]}")

        kwargs = {
            'sender': self.__class__,
            'mail': mail,
            'event': event,
            'message': message,
        }
        signal.send(**kwargs)



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _

__all__ = ['MailConfig']


class MailConfig(AppConfig):
    name = 'django_aws_mail'
    label = 'django_aws_mail'
    verbose_name = _("Django AWS Mail")
    default_auto_field = 'django.db.models.AutoField'

    def ready(self):
        import django_aws_mail.handlers  # noqa



from django.contrib import admin

from django_aws_mail.models import Delay, Bounce, Complaint
from django_aws_mail.forms import DelayForm, BounceForm, ComplaintForm
from django_aws_mail.utils import admin_link


@admin.register(Delay)
class DelayAdmin(admin.ModelAdmin):
    list_display = ('created', 'updated', 'delay_type', 'count', 'destination', 'user_link')
    readonly_fields = ('created',)
    search_fields = ('destination',)
    ordering = ('-updated',)
    form = DelayForm

    list_select_related = ('user',)

    @admin.display(ordering='user__email')
    @admin_link('user', 'User')
    def user_link(self, user):
        return user


@admin.register(Bounce)
class BounceAdmin(admin.ModelAdmin):
    list_display = ('created', 'updated', 'bounce_type', 'bounce_sub_type', 'bounce_status',
                    'count', 'destination', 'user_link')
    readonly_fields = ('created',)
    search_fields = ('destination',)
    ordering = ('-updated',)
    form = BounceForm

    list_select_related = ('user',)

    @admin.display(ordering='bounce__bouncedRecipients__0__status')
    def bounce_status(self, obj):
        try:
            return obj.bounce['bouncedRecipients'][0]['status']
        except KeyError:
            return None

    @admin.display(ordering='user__email')
    @admin_link('user', 'User')
    def user_link(self, user):
        return user


@admin.register(Complaint)
class ComplaintAdmin(admin.ModelAdmin):
    list_display = ('created', 'updated', 'complaint_type', 'complaint_subtype', 'count',
                    'destination', 'user_link')
    readonly_fields = ('created',)
    search_fields = ('destination',)
    ordering = ('-updated',)
    form = ComplaintForm

    list_select_related = ('user',)

    @admin.display(ordering='complaint__complaintSubType')
    def complaint_subtype(self, obj):
        try:
            return obj.complaint['complaintSubType']
        except KeyError:
            return None

    @admin.display(ordering='user__email')
    @admin_link('user', 'User')
    def user_link(self, user):
        return user



from django.contrib import admin
from django.urls import path, include

app_name = 'sandbox'

urlpatterns = [
    path('mail/', include('django_aws_mail.urls', namespace='mail')),
    path('admin/', admin.site.urls),
]



from django.dispatch import Signal

mail_pre_send = Signal()
mail_post_send = Signal()

mail_bounce = Signal()
mail_complaint = Signal()
mail_delivery = Signal()
mail_delivery_delay = Signal()
mail_send = Signal()
mail_reject = Signal()
mail_open = Signal()
mail_click = Signal()



from django import forms

from django_json_widget.widgets import JSONEditorWidget

from django_aws_mail.models import Delay, Bounce, Complaint


class DelayForm(forms.ModelForm):

    class Meta:
        model = Delay
        fields = '__all__'
        widgets = {
            'delay': JSONEditorWidget(width='80%'),
            'mail': JSONEditorWidget(width='80%'),
        }


class BounceForm(forms.ModelForm):

    class Meta:
        model = Bounce
        fields = '__all__'
        widgets = {
            'bounce': JSONEditorWidget(width='80%'),
            'mail': JSONEditorWidget(width='80%'),
        }


class ComplaintForm(forms.ModelForm):

    class Meta:
        model = Complaint
        fields = '__all__'
        widgets = {
            'complaint': JSONEditorWidget(width='80%'),
            'mail': JSONEditorWidget(width='80%'),
        }



from django.urls import reverse
from django.utils.html import format_html

from django_aws_mail.config import mail_settings


def get_mail_type(mail):
    if 'tags' in mail and 'mail-type' in mail['tags']:
        mail_type = mail['tags']['mail-type'][0]
    else:
        mail_type = None

    if mail_type and mail_type in mail_settings.MAIL_TYPES:
        return mail_settings.MAIL_TYPES[mail_type]
    else:
        return 'email of unknown type'


def admin_change_url(obj):
    app_label = obj._meta.app_label
    model_name = obj._meta.model.__name__.lower()
    return reverse('admin:{}_{}_change'.format(app_label, model_name), args=(obj.pk,))


def admin_link(attr, short_description, empty_description="-"):
    """
    Decorator used for rendering a link to a related model in the admin detail page.
    The wrapped method receives the related object and should return the link text.

    Usage:
        @admin_link('credit_card', _('Credit Card'))
        def credit_card_link(self, credit_card):
            return credit_card.name

    :param attr: Name of the related field.
    :param short_description: Name of the field.
    :param empty_description: Value to display if the related field is None.
    """
    def wrap(func):
        def field_func(self, obj):
            related_obj = getattr(obj, attr)
            if related_obj is None:
                return empty_description
            url = admin_change_url(related_obj)
            return format_html('<a href="{}">{}</a>', url, func(self, related_obj))
        field_func.short_description = short_description
        field_func.allow_tags = True
        return field_func
    return wrap



import os
from django.conf import settings


class AppSettings:
    def __init__(self, prefix):
        self.prefix = prefix

    def _get_setting(self, name, default=None):
        full_name = f"{self.prefix}_{name}"

        # 1. Check Django settings
        if hasattr(settings, full_name):
            return getattr(settings, full_name)

        # 2. Check Environment Variables
        val = os.getenv(full_name)
        if val is not None:
            # Convert string representations of booleans
            normalized_val = val.lower().strip()
            if normalized_val in ('true', '1', 'yes', 'on'):
                return True
            if normalized_val in ('false', '0', 'no', 'off'):
                return False
            return val

        # 3. Fallback to Default
        return default

    @property
    def AWS_REGION_NAME(self):
        return self._get_setting('REGION_NAME', 'us-east-1')

    @property
    def AWS_ACCESS_KEY_ID(self):
        return self._get_setting('ACCESS_KEY_ID')

    @property
    def AWS_SECRET_ACCESS_KEY(self):
        return self._get_setting('SECRET_ACCESS_KEY')

    @property
    def AWS_SNS_VERIFY_NOTIFICATION(self):
        return self._get_setting('SNS_VERIFY_NOTIFICATION', True)

    @property
    def AWS_SNS_VERIFY_CERTIFICATE(self):
        return self._get_setting('SNS_VERIFY_CERTIFICATE', True)

    @property
    def AWS_SNS_TOPIC_ARN(self):
        return self._get_setting('SNS_TOPIC_ARN', None)

    @property
    def MAIL_TYPES(self):
        return self._get_setting('TYPES', None)

# Instantiate once
mail_settings = AppSettings('MAIL_AWS')



"""
Sandbox project settings.
"""
from pathlib import Path

DEBUG = True

SECRET_KEY = 'secret'

BASE_DIR = Path(__file__).resolve().parent.parent

ALLOWED_HOSTS = ['*']

CSRF_TRUSTED_ORIGINS = []


INSTALLED_APPS = [
    'sandbox',
    'django_aws_mail',

    # django
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',

    # vendor
    'django_json_widget',
]


MIDDLEWARE = [
    'django.middleware.gzip.GZipMiddleware',
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

ROOT_URLCONF = 'sandbox.urls'

WSGI_APPLICATION = 'sandbox.wsgi.application'


# HTTPS security settings
SECURE_SSL_REDIRECT = False
# SECURE_PROXY_SSL_HEADER = ('HTTP_X_FORWARDED_PROTO', 'https')
# CSRF_COOKIE_SECURE = not DEBUG
# SESSION_COOKIE_SECURE = not DEBUG
# SECURE_CONTENT_TYPE_NOSNIFF = True
# SECURE_BROWSER_XSS_FILTER = True

# SECURE_CROSS_ORIGIN_OPENER_POLICY = 'same-origin-allow-popups'

X_FRAME_OPTIONS = 'DENY'


TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'OPTIONS': {
            'loaders': [
                'django.template.loaders.filesystem.Loader',
                'django.template.loaders.app_directories.Loader',
            ],
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

FORM_RENDERER = 'django.forms.renderers.TemplatesSetting'


# Password validation
# https://docs.djangoproject.com/en/1.10/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
]

AUTH_USER_MODEL = 'auth.User'

AUTHENTICATION_BACKENDS = ['django.contrib.auth.backends.AllowAllUsersModelBackend']


# Internationalization

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)

# STATIC_URL = '/static/'
# MEDIA_URL = '/media/'
#
# STATIC_ROOT = os.getenv('STATIC_LOCATION')
# MEDIA_ROOT = os.getenv('MEDIA_LOCATION')


# Logging

LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'verbose': {
            'format': '%(asctime)s [%(levelname)s] %(pathname)s (%(lineno)d): %(message)s'
        },
    },
    'handlers': {
        'console': {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
            'formatter': 'verbose',
        },
    },
    'loggers': {
        '': {
            'handlers': ['console'],
            'level': 'WARNING',
        },
        'django': {
            'handlers': ['console'],
            'level': 'INFO',
        },
        'django_aws_mail': {
            'handlers': ['console'],
            'level': 'DEBUG',
            'propagate': False,
        }
    },
}


# Database

DEFAULT_AUTO_FIELD = 'django.db.models.AutoField'

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        "NAME": "django_aws_mail.db",
    },
}


# Email

DEFAULT_FROM_EMAIL = '"Example" <noreply@example.com>'
SERVER_EMAIL = '"example" <noreply@example.com>'
ADMINS = ['"example" <admin@example.com>']

# EMAIL_BACKEND = "django.core.mail.backends.smtp.EmailBackend"
EMAIL_BACKEND = 'django_aws_mail.backends.EmailBackend'



"""
Mail backend for handling AWS SES via boto3.
Inspired by django-amazon-ses.
"""
import boto3
from botocore.exceptions import ClientError, BotoCoreError
from email.utils import parseaddr

from django import get_version
from django.core.mail.backends.base import BaseEmailBackend
from django_aws_mail.signals import mail_pre_send, mail_post_send
from django_aws_mail.config import mail_settings


class EmailBackend(BaseEmailBackend):
    """
    An email backend for use with Amazon SESv2.
    https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sesv2.html

    Overrides the default setting:
    EMAIL_BACKEND = 'django.core.mail.backends.smtp.EmailBackend'
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.region_name = kwargs.get('aws_region_name') or mail_settings.AWS_REGION_NAME
        self.access_key_id = kwargs.get('aws_access_key_id') or mail_settings.AWS_ACCESS_KEY_ID
        self.secret_access_key = kwargs.get('aws_secret_access_key') or mail_settings.AWS_SECRET_ACCESS_KEY
        self.connection = None

    def open(self):
        if self.connection:
            return False
        try:
            self.connection = boto3.client(
                'sesv2',
                region_name=self.region_name,
                aws_access_key_id=self.access_key_id,
                aws_secret_access_key=self.secret_access_key
            )
            return True
        except Exception:
            if not self.fail_silently:
                raise
            return False

    def close(self):
        if self.connection:
            self.connection = None

    def _prepare_message(self, email_message):
        """Extracts envelope info and generates raw MIME data."""
        recipients = [parseaddr(addr)[1] for addr in email_message.recipients()]
        _, from_envelope = parseaddr(email_message.from_email)

        if get_version() < '6.0':
            # Support for legacy Django MIME handling
            data = email_message.message().as_bytes(linesep='\r\n')
        else:
            import email.policy
            # Support for modern Django 6.0+ Python Email API
            data = email_message.message(policy=email.policy.SMTPUTF8).as_bytes()

        return from_envelope, recipients, data

    def send_messages(self, email_messages):
        if not email_messages:
            return 0

        self.open()
        if not self.connection:
            return 0

        num_sent = 0
        for msg in email_messages:
            if self._send(msg):
                num_sent += 1

        return num_sent

    def _send(self, email_message):
        mail_pre_send.send(self.__class__, message=email_message)
        if not email_message.recipients():
            return False

        from_envelope, recipients, data = self._prepare_message(email_message)

        try:
            response = self.connection.send_email(
                FromEmailAddress=from_envelope,
                Destination={'ToAddresses': recipients},
                Content={'Raw': {'Data': data}}
            )
        except (BotoCoreError, ClientError):
            if not self.fail_silently:
                raise
            return False

        mail_post_send.send(self.__class__, message=email_message, response=response)
        return True



import logging

from django.dispatch import receiver
from django.contrib.auth import get_user_model

from django_aws_mail.signals import mail_delivery, mail_delivery_delay, mail_bounce, mail_complaint
from django_aws_mail.models import Delay, Bounce, Complaint

User = get_user_model()

logger = logging.getLogger(__name__)


@receiver(mail_delivery)
def handle_delivery(sender, mail, event, message, **kwargs):
    logger.debug(f'Message delivered:\n{message}\n')


@receiver(mail_delivery_delay)
def handle_delay(sender, mail, event, message, **kwargs):
    logger.debug(f'Message delayed:\n{message}\n')

    email = mail['destination'][0]
    try:
        user = User.objects.get(email=email.lower())
    except User.DoesNotExist:
        user = None

    # log delay
    obj, created = Delay.objects.get_or_create(
        destination=email,
        delay_type=event.get('delayType'),
        defaults={'user': user, 'delay': event, 'mail': mail})

    if not created:
        obj.user = user
        obj.delay = event
        obj.mail = mail
        obj.count += 1
        obj.save()


@receiver(mail_bounce)
def handle_bounce(sender, mail, event, message, **kwargs):
    logger.debug(f'Message bounced:\n{message}\n')

    email = mail['destination'][0]
    try:
        user = User.objects.get(email=email.lower())
    except User.DoesNotExist:
        user = None

    # log bounce
    obj, created = Bounce.objects.get_or_create(
        destination=email,
        bounce_type=event.get('bounceType'),
        bounce_sub_type=event.get('bounceSubType'),
        defaults={'user': user, 'bounce': event, 'mail': mail})

    if not created:
        obj.user = user
        obj.bounce = event
        obj.mail = mail
        obj.count += 1
        obj.save()


@receiver(mail_complaint)
def handle_complaint(sender, mail, event, message, **kwargs):
    logger.debug(f'Message complaint:\n{message}\n')

    email = mail['destination'][0]
    try:
        user = User.objects.get(email=email.lower())
    except User.DoesNotExist:
        user = None

    # log complaint
    obj, created = Complaint.objects.get_or_create(
        destination=email,
        complaint_type=event.get('complaintSubType'),
        defaults={'user': user, 'complaint': event, 'mail': mail})

    if not created:
        obj.user = user
        obj.complaint = event
        obj.mail = mail
        obj.count += 1
        obj.save()



from email.utils import formataddr

from django.conf import settings
from django.core.mail import EmailMultiAlternatives
from django.template.loader import render_to_string
from django.utils.html import strip_spaces_between_tags

from django_aws_mail.html import HTMLParser


def compose(recipients, subject, template, context=None, from_email=None, **kwargs):
    """
    Create a multipart MIME email message, by rendering html and text body.
    Optionally add SES headers config_set and mail_type to the kwargs.
    """
    # sanitize input: subject, recipients, from email
    subject = ''.join(subject.splitlines())
    if not isinstance(recipients, list):
        recipients = [recipients]
    if not from_email:
        from_email = settings.DEFAULT_FROM_EMAIL
    if not isinstance(from_email, str):
        from_email = formataddr(from_email)

    # render html content
    context = context or {}
    html = render_to_string(template, context).strip()
    html = strip_spaces_between_tags(html)

    # convert html to text
    parser = HTMLParser()
    parser.feed(html)
    parser.close()
    text = parser.text()

    # get optional headers
    headers = {}
    if 'headers' in kwargs:
        headers = kwargs.pop('headers')
    if 'config_set' in kwargs:
        headers['X-Ses-Configuration-Set'] = kwargs.pop('config_set')
    if 'mail_type' in kwargs:
        headers['X-Ses-Message-Tags'] = f"mail-type={kwargs.pop('mail_type')}"

    # create email message
    message = EmailMultiAlternatives(subject, text, from_email, recipients, headers=headers, **kwargs)
    message.attach_alternative(html, 'text/html')
    return message



from django.db import models
from django.conf import settings
from django.utils import timezone
from django.utils.html import urlize, linebreaks
from django.utils.translation import gettext_lazy as _

from django_aws_mail.utils import get_mail_type

FEEDBACK_BOUNCE_MESSAGE = {
    'General':
        _('The {mail_type} for <b>{email_address}</b> could not be delivered because '
          'of general issues. See details.'),
    'NoEmail':
        _('The {mail_type} for <b>{email_address}</b> bounced because the provided '
          'email address does not appear to exist.'),
    'Suppressed':
        _('The {mail_type} for <b>{email_address}</b> bounced due to previous sending issues.'),
    'OnAccountSuppressionList':
        _('The {mail_type} for <b>{email_address}</b> bounced due to previous sending issues.'),
    'MailboxFull':
        _('The {mail_type} for <b>{email_address}</b> bounced because your inbox '
          'appears to be full.'),
    'ContentRejected':
        _('The {mail_type} for <b>{email_address}</b> was rejected due to suspected SPAM.'),
}

FEEDBACK_DELAY_MESSAGE = {
    'InternalFailure':
        _('The {mail_type} for <b>{email_address}</b> was delayed due to an internal '
          'failure at the MTA.'),
    'General':
        _('The {mail_type} for <b>{email_address}</b> was delayed because there was a '
          'failure of generic nature.'),
    'MailboxFull':
        _('The {mail_type} for <b>{email_address}</b> was delayed because your mailbox '
          'appears to be full and unable to receive additional messages. Make some space!'),
    'SpamDetected':
        _('The {mail_type} for <b>{email_address}</b> was delayed because your mail provider '
          'considers our email to be SPAM.'),
    'RecipientServerError':
        _('The {mail_type} for <b>{email_address}</b> was delayed because there seems '
          'to be a temporary issue at your mail provider.'),
    'IPFailure':
        _('The {mail_type} for <b>{email_address}</b> was delayed because your mail provider '
          'is blocking or throttling mail from our MTA.'),
    'TransientCommunicationFailure':
        _('The {mail_type} for <b>{email_address}</b> was delayed because there was a temporary '
          'communications failure with your mail provider.'),
    'TransientCommunicationGeneral':
        _('The {mail_type} for <b>{email_address}</b> was delayed because there was a temporary '
          'communications failure with your mail provider.'),
}

# https://www.iana.org/assignments/smtp-enhanced-status-codes/smtp-enhanced-status-codes.xhtml
SMTP_STATUS_CODES = {
    '4.4.2': _('Delivery temporarily suspended.'),
    '4.4.7': _('Message expired, mailbox full.'),
    '5.1.0': _('Sender rejected.'),
    '5.1.1': _('Email account does not exist. Probably a typo in the email address before the @.'),
    '5.1.2': _('Unknown mail server.'),
    '5.2.0': _('Mail rejected.'),
    '5.2.1': _('Addressee unknown.'),
    '5.2.2': _('Mailbox full.'),
    '5.3.0': _('Recipient does not exist, invalid mailbox, unknown user.'),
    '5.4.4': _('Invalid domain. Probably a typo in the email address after the @.'),
    '5.5.0': _('Mailbox unavailable.'),
    '5.7.1': _('Message blocked due to spam.'),
}


class Delay(models.Model):
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        null=True,
        related_name='email_delays')

    destination = models.EmailField()

    delay_type = models.CharField(max_length=50, blank=True, null=True)

    delay = models.JSONField()
    mail = models.JSONField()

    updated = models.DateTimeField(_('Date updated'), auto_now=True)
    created = models.DateTimeField(_('Date created'), default=timezone.now)

    count = models.IntegerField(default=1)

    class Meta:
        ordering = ('-created',)
        constraints = [
            models.UniqueConstraint(fields=['destination', 'delay_type'],
                                    name='unique_delay'),
        ]

    def __str__(self):
        return (f"Mail Delay for "
                f"{self.delay['delayedRecipients'][0]['emailAddress']} ({self.count})")

    def get_message(self):
        delay_type = self.delay['delayType']
        if delay_type in FEEDBACK_DELAY_MESSAGE:
            message = FEEDBACK_DELAY_MESSAGE[delay_type]
        else:
            message = 'The {mail_type} was delayed for undetermined reasons.'

        # add mail type and email address to message
        message = message.format(
            mail_type=get_mail_type(self.mail),
            email_address=self.mail['destination'][0])

        status_code = None
        status_message = None
        if 'delayedRecipients' in self.delay and 'status' in self.delay['delayedRecipients'][0]:
            status_code = self.delay['delayedRecipients'][0]['status']
            if status_code in SMTP_STATUS_CODES:
                status_message = SMTP_STATUS_CODES[status_code]

        if status_code and status_message:
            message = (f'{message} Additionally reported status code '
                       f'{status_code}: {status_message}')

        return message

    def get_diagnostics(self):
        diagnostic_message = None
        if ('delayedRecipients' in self.delay and
                'diagnosticCode' in self.delay['delayedRecipients'][0]):
            diagnostic_code = self.delay['delayedRecipients'][0]['diagnosticCode']
            diagnostic_message = urlize(linebreaks(diagnostic_code.strip()))
        return diagnostic_message


class Bounce(models.Model):
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        null=True,
        related_name='email_bounces')

    destination = models.EmailField()

    bounce_type = models.CharField(max_length=50, blank=True, null=True)
    bounce_sub_type = models.CharField(max_length=50, blank=True, null=True)

    bounce = models.JSONField()
    mail = models.JSONField()

    updated = models.DateTimeField(_('Date updated'), auto_now=True)
    created = models.DateTimeField(_('Date created'), default=timezone.now)

    count = models.IntegerField(default=1)

    class Meta:
        ordering = ('-created',)
        constraints = [
            models.UniqueConstraint(fields=['destination', 'bounce_type', 'bounce_sub_type'],
                                    name='unique_bounce'),
        ]

    def __str__(self):
        return (f"{self.bounce_type} Mail Bounce for "
                f"{self.bounce['bouncedRecipients'][0]['emailAddress']} ({self.count})")

    def get_message(self):
        bounce_sub_type = self.bounce['bounceSubType']
        if bounce_sub_type in FEEDBACK_BOUNCE_MESSAGE:
            message = FEEDBACK_BOUNCE_MESSAGE[bounce_sub_type]
        else:
            message = 'The {mail_type} bounced for undetermined reasons.'

        # add mail type and email address to message
        message = message.format(
            mail_type=get_mail_type(self.mail),
            email_address=self.mail['destination'][0])

        status_code = None
        status_message = None
        if 'bouncedRecipients' in self.bounce and 'status' in self.bounce['bouncedRecipients'][0]:
            status_code = self.bounce['bouncedRecipients'][0]['status']
            if status_code in SMTP_STATUS_CODES:
                status_message = SMTP_STATUS_CODES[status_code]

        if status_code and status_message:
            message = (f'{message} Additionally reported status code '
                       f'{status_code}: {status_message}')

        return message

    def get_diagnostics(self):
        diagnostic_message = None
        if ('bouncedRecipients' in self.bounce and
                'diagnosticCode' in self.bounce['bouncedRecipients'][0]):
            diagnostic_code = self.bounce['bouncedRecipients'][0]['diagnosticCode']
            diagnostic_message = urlize(linebreaks(diagnostic_code.strip()))

        return diagnostic_message


class Complaint(models.Model):
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        null=True,
        related_name='email_complaints')

    destination = models.EmailField()

    complaint_type = models.CharField(max_length=50, blank=True, null=True)

    complaint = models.JSONField()
    mail = models.JSONField()

    updated = models.DateTimeField(_('Date updated'), auto_now=True)
    created = models.DateTimeField(_('Date created'), default=timezone.now)

    count = models.IntegerField(default=1)

    class Meta:
        ordering = ('-created',)
        constraints = [
            models.UniqueConstraint(fields=['destination', 'complaint_type'],
                                    name='unique_complaint'),
        ]

    def __str__(self):
        return (f"Mail Complaint for "
                f"{self.complaint['complainedRecipients'][0]['emailAddress']} ({self.count})")

    def get_message(self):
        message = (
            'The {mail_type} for <b>{email_address}</b> could not be delivered because it '
            'bounced with a complaint, probably due to SPAM rules at your mail provider.')

        # add mail type and email address to message
        message = message.format(
            mail_type=get_mail_type(self.mail),
            email_address=self.mail['destination'][0])

        return message
